use rayon::prelude::*;
use std::collections::HashSet;
use std::fs;
use std::sync::{Arc, Mutex};

pub fn solve() {
    let path = "res/05/input";

    let contents: Vec<String> = fs::read_to_string(path)
        .expect("I/O error, wrong path?")
        .split("\n\n")
        .map(|s| String::from(s))
        .collect();
    //let contents = BufReader::new(fs::File::open(path).expect("I/O error, wrong path?"));

    let seeds_raw: Vec<usize> = contents[0]
        .split_whitespace()
        .skip(1)
        .map(|s| s.parse::<usize>().unwrap())
        .collect();
    let seeds = seeds_raw.clone();

    let seeds_len = seeds.len();
    let seeds_arc = Arc::new(Mutex::new(seeds));

    parse_maps(&seeds_arc, seeds_len, &contents);

    let seeds = seeds_arc.lock().unwrap();
    let result: usize = *seeds.iter().min().unwrap();

    println!("Result 1: {}", result);

    let maps_ranges: Vec<Vec<Vec<usize>>> = contents
        .iter()
        .skip(1)
        .map(|s| {
            s.lines()
                .skip(1)
                .map(|s| {
                    s.split_whitespace()
                        .map(|s| s.parse::<usize>().unwrap())
                        .collect::<Vec<usize>>()
                })
                .collect::<Vec<Vec<usize>>>()
        })
        .collect();

    let mut current: HashSet<(usize, usize)> = seeds_raw
        .chunks(2)
        .map(|arr| (arr[0], arr[0] + arr[1] - 1))
        .collect();
    let mut gen: HashSet<(usize, usize)>;

    for ranges in maps_ranges.iter() {
        let next = ranges.clone();

        gen = HashSet::<(usize, usize)>::new();

        let mut not_yet_mapped = current.clone();
        for map in next {
            let mut tmp = HashSet::<(usize, usize)>::new();
            for (s, e) in not_yet_mapped.clone().iter() {
                if map[1] < *s && map[1] + map[2] >= *s && map[1] + map[2] <= *e {
                    // left
                    let len = map[1] + map[2] - *s;
                    let dest = map[0] + (*s - map[1]);

                    gen.insert((dest, dest + len));
                    tmp.insert((*s + len, *e));
                } else if map[1] >= *s && map[1] <= *e && map[1] + map[2] > *e {
                    // right
                    let len = *e - map[1];
                    let dest = map[0];

                    gen.insert((dest, dest + len));
                    tmp.insert((*s, *e - len));
                } else if map[1] >= *s && map[1] + map[2] <= *e {
                    // inner
                    let len = map[2];
                    let dest = map[0];

                    gen.insert((dest, dest + len));
                    tmp.insert((*s, map[1]));
                    tmp.insert((map[1] + map[2], *e));
                } else if map[1] < *s && map[1] + map[2] > *e {
                    // outer
                    let len = *e - *s;
                    let dest = map[0] + (*s - map[1]);

                    gen.insert((dest, dest + len));
                } else {
                    tmp.insert((*s, *e));
                }
            }
            not_yet_mapped = tmp;
        }
        not_yet_mapped.iter().for_each(|v| {
            gen.insert(*v);
        });
        current = gen;
    }

    let mut arr = current
        .iter()
        .filter(|(s, e)| *s != *e)
        .map(|(s, _)| *s)
        .collect::<Vec<usize>>();
    arr.sort();

    println!("Result 2: {}", arr[0]);
}

fn parse_maps(seeds_arc: &Arc<Mutex<Vec<usize>>>, seeds_len: usize, contents: &Vec<String>) {
    contents.iter().skip(1).for_each(|s| {
        let ranges: Vec<Vec<usize>> = s
            .lines()
            .skip(1)
            .map(|s| {
                s.split_whitespace()
                    .map(|s| s.parse::<usize>().unwrap())
                    .collect::<Vec<usize>>()
            })
            .collect();

        (0..seeds_len).par_bridge().for_each(|i| {
            for j in 0..ranges.len() {
                let range_start = ranges[j][1];
                let range_end = ranges[j][1] + ranges[j][2];
                let mut seeds = seeds_arc.lock().unwrap();
                if seeds[i] >= range_start && seeds[i] <= range_end {
                    seeds[i] = ranges[j][0] + (seeds[i] - range_start);
                    break;
                }
            }
        })
    })
}
