use std::{collections::HashMap, fs};

use rayon::prelude::*;

pub fn solve() {
    let path = "res/12/input";

    let contents = fs::read_to_string(path).expect("I/O error, wrong path?");
    //let contents = BufReader::new(fs::File::open(path).expect("I/O error, wrong path?"));

    let result: usize = contents
        .par_lines()
        .map(|line| {
            let (springs_str, nums_str) = line.split_once(" ").unwrap();
            let nums: Vec<usize> = nums_str
                .split(",")
                .map(|n| n.parse::<usize>().unwrap())
                .collect();

            let mut springs: Vec<char> = springs_str.chars().collect();
            springs.push('.');

            let mut cache = HashMap::new();

            count_spring_arrangements(&mut cache, springs_str.as_bytes(), None, &nums)
        })
        .sum();

    println!("Result 1: {result}");

    let result: usize = contents
        .par_lines()
        .map(|line| {
            let (springs_str, nums_str) = line.split_once(" ").unwrap();
            let nums_tmp: Vec<usize> = nums_str
                .split(",")
                .map(|n| n.parse::<usize>().unwrap())
                .collect();

            let nums: Vec<usize> = nums_tmp
                .iter()
                .cycle()
                .take(nums_tmp.len() * 5)
                .map(|i| *i)
                .collect();
            let springs: String = std::iter::repeat(String::from(springs_str))
                .take(5)
                .collect::<Vec<String>>()
                .join("?");

            let mut cache = HashMap::new();

            count_spring_arrangements(&mut cache, springs.as_bytes(), None, &nums)
        })
        .sum();

    println!("Result 2: {result}");
}

fn count_spring_arrangements(
    cache: &mut HashMap<(usize, usize, usize), usize>,
    springs: &[u8],
    in_broken_group: Option<usize>,
    remaining_broken: &[usize],
) -> usize {
    if springs.is_empty() {
        return match (in_broken_group, remaining_broken.len()) {
            (None, 0) => 1,
            (Some(x), 1) if x == remaining_broken[0] => 1,
            _ => 0,
        };
    }

    if in_broken_group.is_some() && remaining_broken.is_empty() {
        return 0;
    }

    let key = (
        springs.len(),
        in_broken_group.unwrap_or(0),
        remaining_broken.len(),
    );
    if let Some(&result) = cache.get(&key) {
        return result;
    }

    let ways = match (springs[0], in_broken_group) {
        (b'.', Some(x)) if x != remaining_broken[0] => 0,
        (b'.', Some(_)) => {
            count_spring_arrangements(cache, &springs[1..], None, &remaining_broken[1..])
        }
        (b'.', None) => count_spring_arrangements(cache, &springs[1..], None, remaining_broken),
        (b'#', Some(_)) => count_spring_arrangements(
            cache,
            &springs[1..],
            in_broken_group.map(|x| x + 1),
            remaining_broken,
        ),
        (b'#', None) => count_spring_arrangements(cache, &springs[1..], Some(1), remaining_broken),
        (b'?', Some(x)) => {
            let mut ans = count_spring_arrangements(
                cache,
                &springs[1..],
                in_broken_group.map(|x| x + 1),
                remaining_broken,
            );
            if x == remaining_broken[0] {
                ans +=
                    count_spring_arrangements(cache, &springs[1..], None, &remaining_broken[1..]);
            }
            ans
        }
        (b'?', None) => {
            count_spring_arrangements(cache, &springs[1..], Some(1), remaining_broken)
                + count_spring_arrangements(cache, &springs[1..], None, remaining_broken)
        }
        _ => unreachable!(),
    };

    cache.insert(key, ways);
    ways
}
