use std::{
    cmp::Reverse,
    collections::{BinaryHeap, HashMap, HashSet},
    fs,
};

pub fn solve() {
    let path = "res/17/input";

    let contents = fs::read_to_string(path).expect("I/O error, wrong path?");
    //let contents = BufReader::new(fs::File::open(path).expect("I/O error, wrong path?"));

    let grid: Vec<Vec<u32>> = contents
        .lines()
        .map(|line| {
            line.chars()
                .map(|c| c.to_digit(10).unwrap())
                .collect::<Vec<u32>>()
        })
        .collect();

    let start = (0, 0);
    let target = (grid.len() - 1, grid[0].len() - 1);

    let (result1, result2) = rayon::join(
        || dijkstra(&grid, start, target, 0, 3),
        || dijkstra(&grid, start, target, 4, 10),
    );
    println!("Result 1: {result1}");
    println!("Result 2: {result2}");
}

fn dijkstra(
    graph: &Vec<Vec<u32>>,
    start: (usize, usize),
    goal: (usize, usize),
    min_steps: usize,
    max_steps: usize,
) -> u32 {
    let mut queue = BinaryHeap::new();
    let mut visited = HashSet::new();
    let mut dist = HashMap::new();
    let mut queue_invalid = HashSet::new();

    queue.push(Reverse((0, start, start, 0, 0)));
    queue.push(Reverse((0, start, start, 1, 0)));

    dist.insert((start, 0, 0), 0);
    dist.insert((start, 1, 0), 0);

    while let Some(Reverse((d, pos, last_pos, dir, steps))) = queue.pop() {
        if visited.contains(&(pos, dir, steps)) {
            continue;
        }

        if queue_invalid.contains(&(pos, d)) {
            queue_invalid.remove(&(pos, d));
            continue;
        }

        visited.insert((pos, dir, steps));

        let adj_list: Vec<((usize, usize), u32, i32, i32)> =
            dijkstra_adj_pos(pos, last_pos, dir, steps, min_steps, max_steps)
                .iter()
                .filter(|((x, y), _, _)| *x < graph.len() && *y < graph[0].len())
                .map(|(pos, dir, steps)| (*pos, graph[pos.0][pos.1], *dir, *steps))
                .collect();

        for (next, weight, next_dir, next_steps) in adj_list {
            if !dist.contains_key(&(next, next_dir, next_steps))
                || dist[&(pos, dir, steps)] + weight < dist[&(next, next_dir, next_steps)]
            {
                if dist.contains_key(&(next, next_dir, next_steps)) {
                    queue_invalid.insert((next, dist[&(next, next_dir, next_steps)]));
                }

                dist.insert(
                    (next, next_dir, next_steps),
                    dist[&(pos, dir, steps)] + weight,
                );
                queue.push(Reverse((
                    dist[&(pos, dir, steps)] + weight,
                    next,
                    pos,
                    next_dir,
                    next_steps,
                )));
            }
        }
    }
    *dist
        .iter()
        .filter(|((pos, _, _), _)| *pos == goal)
        .map(|(_, v)| v)
        .min()
        .unwrap()
}

fn dijkstra_adj_pos(
    pos: (usize, usize),
    last_pos: (usize, usize),
    dir: i32,
    steps: i32,
    min_steps: usize,
    max_steps: usize,
) -> Vec<((usize, usize), i32, i32)> {
    let mut adj_positions = Vec::new();

    let directions = vec![(0, 1), (1, 0), (0, -1), (-1, 0)]; // Right, Down, Left, Up

    for &(dx, dy) in &directions {
        let new_pos = (pos.0 as isize + dx, pos.1 as isize + dy);

        if new_pos.0 < 0 || new_pos.1 < 0 {
            continue;
        }

        let new_pos = (new_pos.0 as usize, new_pos.1 as usize);

        if new_pos == last_pos {
            // wrong dir
            continue;
        }

        if new_pos.0.abs_diff(last_pos.0) == 2
            || new_pos.1.abs_diff(last_pos.1) == 2
            || pos == last_pos
        {
            if steps < max_steps as i32 {
                adj_positions.push((new_pos, dir, steps + 1))
            }
        } else if steps >= min_steps as i32 {
            let dir = match (dx, dy) {
                (0, 1) => 0,
                (1, 0) => 1,
                (0, -1) => 2,
                _ => 3,
            };
            adj_positions.push((new_pos, dir, 1));
        }
    }

    adj_positions
}
