use std::fs;

pub fn solve() {
    let path = "res/18/input";

    let contents = fs::read_to_string(path).expect("I/O error, wrong path?");
    //let contents = BufReader::new(fs::File::open(path).expect("I/O error, wrong path?"));

    let mut curr = (0isize, 0isize);

    let mut c_list = vec![curr];
    let mut edge = 0;

    contents.lines().for_each(|line| {
        let parts: Vec<&str> = line.split_whitespace().collect();
        let dir = parts[0];
        let steps: isize = parts[1].parse().unwrap();

        edge += steps;
        curr = match dir {
            "U" => (curr.0 - steps, curr.1),
            "R" => (curr.0, curr.1 + steps),
            "D" => (curr.0 + steps, curr.1),
            "L" => (curr.0, curr.1 - steps),
            _ => panic!(),
        };
        c_list.push(curr);
    });

    // Shoelace Formula
    let result: isize = c_list
        .iter()
        .zip(c_list.iter().skip(1))
        .map(|((x1, y1), (x2, y2))| *x1 * *y2 - *x2 * *y1)
        .sum::<isize>()
        / 2;

    let result = (result.abs() as f64 - 0.5 * (edge as f64) + 1.0) + edge as f64;

    println!("Result 1: {result}");

    let mut curr = (0isize, 0isize);

    let mut c_list = vec![curr];
    let mut edge = 0;
    contents.lines().for_each(|line| {
        let (_, hex) = line.split_once('#').unwrap();
        let dir = isize::from_str_radix(&hex[5..6], 16).unwrap();
        let steps = isize::from_str_radix(&hex[0..5], 16).unwrap();

        edge += steps;
        curr = match dir {
            3 => (curr.0 - steps, curr.1),
            0 => (curr.0, curr.1 + steps),
            1 => (curr.0 + steps, curr.1),
            2 => (curr.0, curr.1 - steps),
            _ => panic!(),
        };
        c_list.push(curr);
    });

    // Shoelace Formula
    let result: isize = c_list
        .iter()
        .zip(c_list.iter().skip(1))
        .map(|((x1, y1), (x2, y2))| *x1 * *y2 - *x2 * *y1)
        .sum::<isize>()
        / 2;

    let result = (result.abs() as f64 - 0.5 * (edge as f64) + 1.0) + edge as f64;

    println!("Result 2: {result}");
}
